"""
This linter checks for too high offset SELECT queries
"""
from collections import OrderedDict

from sql_metadata.compat import get_query_limit_and_offset, get_query_tables

from indexdigest.utils import LinterEntry, shorten_query, is_select_query


OFFSET_THRESHOLD = 1000


def check_high_offset_selects(_, queries):
    """
    :type _ indexdigest.database.Database
    :type queries list[str]
    :rtype: list[LinterEntry]
    """
    for query in queries:
        # ignore insert queries (#140)
        if not is_select_query(query):
            continue

        res = get_query_limit_and_offset(query)

        if res is None:
            continue

        (limit, offset) = res

        if offset < OFFSET_THRESHOLD:
            continue

        table_name = get_query_tables(query)[0]

        context = OrderedDict()
        context['query'] = query
        context['limit'] = limit
        context['offset'] = offset

        yield LinterEntry(linter_type='high_offset_selects', table_name=table_name,
                          message='"{}" query uses too high offset impacting the performance'.
                          format(shorten_query(query)),
                          context=context)



"""
This linter checks for databases with no tables
"""
from indexdigest.utils import LinterEntry


def get_empty_databases(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[str]
    """
    for db_name in database.query_list('SHOW DATABASES'):
        # skip "core" MySQL databases
        if db_name in ['information_schema']:
            continue

        tables_count = database.query_field('SELECT COUNT(*) FROM information_schema.TABLES '
                                            'WHERE TABLE_SCHEMA = "{}" AND '
                                            'TABLE_TYPE = "BASE TABLE"'.format(db_name))
        # print(db_name, tables_count)
        if tables_count == 0:
            yield db_name


def check_empty_database(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    for db_name in get_empty_databases(database):
        yield LinterEntry(linter_type='empty_database', table_name=db_name,
                          message='"{}" database has no tables'.format(db_name))



"""
This linter looks for table that were not updated recently
"""
from collections import OrderedDict
from datetime import datetime
from time import time

from indexdigest.utils import LinterEntry

from .linter_0028_data_too_old import get_time_columns, get_boundary_times


def check_data_not_updated_recently(database, env=None):
    """
    :type database  indexdigest.database.Database
    :type env dict
    :rtype: list[LinterEntry]
    """
    now = int(time())  # I will probably never understand dates handling in Python

    # set up a diff threshold (in days)
    env = env if env else dict()
    diff_threshold = int(env.get('INDEX_DIGEST_DATA_NOT_UPDATED_RECENTLY_THRESHOLD_DAYS', 30))

    for (table_name, column) in get_time_columns(database):
        timestamps = get_boundary_times(database, table_name, column)
        if timestamps is None:
            continue

        diff = now - timestamps.get('max')
        # print(table_name, column, timestamps, now, diff)

        if diff > diff_threshold * 86400:
            diff_days = int(diff / 86400)

            metadata = database.get_table_metadata(table_name)

            context = OrderedDict()
            context['diff_days'] = diff_days
            context['data_since'] = str(datetime.fromtimestamp(timestamps.get('min')))
            context['data_until'] = str(datetime.fromtimestamp(timestamps.get('max')))
            context['date_column_name'] = str(column)
            context['schema'] = database.get_table_schema(table_name)
            context['rows'] = database.get_table_rows_estimate(table_name)
            context['table_size_mb'] = \
                1. * (metadata['data_size'] + metadata['index_size']) / 1024 / 1024

            yield LinterEntry(linter_type='data_not_updated_recently', table_name=table_name,
                              message='"{}" has the latest row added {} days ago, '
                                      'consider checking if it should be up-to-date'.
                              format(table_name, diff_days),
                              context=context)



"""
This linter checks for ...
"""
from collections import OrderedDict

from indexdigest.utils import LinterEntry

# skip small tables
ROWS_COUNT_THRESHOLD = 100000

# cardinality threshold
INDEX_CARDINALITY_THRESHOLD = 6

# the least frequent value should be used at most by x% rows
INDEX_VALUE_PERCENTAGE_THRESHOLD = 20


def get_low_cardinality_indices(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list
    """
    for table_name in database.get_tables():
        rows_count = database.get_table_rows_estimate(table_name)
        if rows_count < ROWS_COUNT_THRESHOLD:
            continue

        # get table indices statistics
        # @see https://dev.mysql.com/doc/refman/5.7/en/show-index.html
        # @see https://www.percona.com/blog/2007/08/28/do-you-always-need-index-on-where-column/
        indices = database.query_dict_rows(
            "select TABLE_NAME, INDEX_NAME, COLUMN_NAME, CARDINALITY from"
            " INFORMATION_SCHEMA.STATISTICS where"
            " TABLE_NAME = '{table_name}' AND TABLE_SCHEMA = '{database_name}'".format(
                table_name=table_name, database_name=database.db_name)
        )

        for index in indices:
            # the cardinality is too high
            if index['CARDINALITY'] > INDEX_CARDINALITY_THRESHOLD:
                continue

            yield table_name, rows_count, index


def check_low_cardinality_index(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    for table_name, rows_count, index in get_low_cardinality_indices(database):
        # the least frequent value should be used in up to 20% of rows
        # https://www.percona.com/blog/2007/08/28/do-you-always-need-index-on-where-column/
        row = database.query_dict_row(
            'SELECT {column} AS value, COUNT(*) AS cnt FROM `{table}` '
            'GROUP BY 1 ORDER BY 2 ASC LIMIT 1'.format(
                column=index['COLUMN_NAME'], table=index['TABLE_NAME']
            )
        )

        value_usage = 100. * row['cnt'] / rows_count
        # print(row, value_usage)

        # the least frequent value is quite rare - it makes sense to have an index here
        if value_usage < INDEX_VALUE_PERCENTAGE_THRESHOLD:
            continue

        # print(value_usage, index, table_name)

        context = OrderedDict()
        context['column_name'] = index['COLUMN_NAME']
        context['index_name'] = index['INDEX_NAME']
        context['index_cardinality'] = int(index['CARDINALITY'])
        context['schema'] = database.get_table_schema(table_name)
        context['value_usage'] = value_usage

        yield LinterEntry(linter_type='low_cardinality_index', table_name=table_name,
                          message='"{}" index on "{}" column has low cardinality, '
                                  'check if it is needed'.
                          format(index['INDEX_NAME'], index['COLUMN_NAME']),
                          context=context)



# pylint: disable=line-too-long
"""index_digest

Analyses your database queries and schema and suggests indices improvements.

Usage:
  index_digest DSN [--sql-log=<file>] [--format=<formatter>] [--analyze-data] [--check-empty-databases] [--checks=<checks> | --skip-checks=<skip-checks>] [--tables=<tables> | --skip-tables=<skip-tables>]
  index_digest (-h | --help)
  index_digest --version

Options:
  DSN               Data Source Name of database to check
  --sql-log=<file>  Text file with SQL queries to check against the database
  --format=<formatter>  Use a given results formatter (plain, syslog, yaml)
  --analyze-data    Run additional checks that will query table data (can be slow!)
  --check-empty-databases  Detect empty databases on this MySQL server
  --checks=<list>   Comma-separated lists of checks to report
  --skip-checks=<list> Comma-separated lists of checks to skip from report
  --tables=<list>   Comma-separated lists of tables to report
  --skip-tables=<list> Comma-separated lists of tables to skip from report
  -h --help         Show this screen.
  --version         Show version.

Examples:
  index_digest mysql://username:password@localhost/dbname
  index_digest mysql://index_digest:qwerty@localhost/index_digest --sql-log=sql.log
  index_digest mysql://index_digest:qwerty@localhost/index_digest --skip-checks=non_utf_columns
  index_digest mysql://index_digest:qwerty@localhost/index_digest --analyze-data --checks=data_too_old,data_not_updated_recently
  index_digest mysql://index_digest:qwerty@localhost/index_digest --analyze-data --skip-tables=DATABASECHANGELOG,DATABASECHANGELOGLOCK

Visit <https://github.com/macbre/index-digest>
"""
from __future__ import print_function

import logging
from itertools import chain
from os import getenv, environ

from docopt import docopt

import indexdigest
from indexdigest.database import Database
from indexdigest.utils import IndexDigestError
from indexdigest.formatters import \
    format_plain, \
    format_syslog, \
    format_yaml
from indexdigest.linters import \
    check_queries_using_filesort, check_queries_using_temporary, \
    check_not_used_indices, check_queries_not_using_indices, \
    check_not_used_tables, check_not_used_columns, \
    check_redundant_indices, \
    check_full_table_scan, \
    check_latin_columns, \
    check_selects_with_like, \
    check_missing_primary_index, \
    check_test_tables, \
    check_insert_ignore_queries, \
    check_single_column, \
    check_empty_tables, \
    check_select_star, \
    check_having_clause, \
    check_data_too_old, \
    check_data_not_updated_recently, \
    check_generic_primary_key, \
    check_high_offset_selects, \
    check_use_innodb, \
    check_empty_database, \
    check_low_cardinality_index


def get_version() -> str:
    """
    Return version string, e.g.
    1.5.0 (git 1a258db)
    """
    return '{version} (git {commit})'.format(
        version=indexdigest.VERSION,
        commit=getenv('COMMIT_SHA', 'dev')[:7]
    )


def get_reports(database, sql_log=None, analyze_data=False, check_empty_databases=False):
    """
    :type database Database
    :type sql_log str
    :type analyze_data bool
    :type check_empty_databases bool
    :rtype: list[indexdigest.utils.LinterEntry]
    """
    logger = logging.getLogger(__name__)

    # read SQL log file (if provided)
    if sql_log:
        logger.debug('Trying to open SQL log file: %s', sql_log)

        with open(sql_log, encoding='utf-8') as log_file:
            queries = log_file.readlines()
            queries = list(map(str.strip, queries))  # remove trailing spaces
            logger.debug('Got %d entries in SQL log file', len(queries))
    else:
        queries = None

    # run all checks
    reports = chain(
        check_redundant_indices(database),
        check_latin_columns(database),
        check_missing_primary_index(database),
        check_test_tables(database),
        check_single_column(database),
        check_empty_tables(database),
        check_generic_primary_key(database),
        check_use_innodb(database),
        check_low_cardinality_index(database),
    )

    # checks that use SQL log
    if queries:
        reports = chain(
            reports,
            check_not_used_indices(database, queries=queries),
            check_not_used_tables(database, queries=queries),
            check_not_used_columns(database, queries=queries),
            check_queries_not_using_indices(database, queries=queries),
            check_queries_using_filesort(database, queries=queries),
            check_queries_using_temporary(database, queries=queries),
            check_full_table_scan(database, queries=queries),
            check_selects_with_like(database, queries=queries),
            check_insert_ignore_queries(database, queries=queries),
            check_select_star(database, queries=queries),
            check_having_clause(database, queries=queries),
            check_high_offset_selects(database, queries=queries),
        )

    # checks that require --analyze-data switch to be on (see #28)
    if analyze_data is True:
        logger.info("Will run data analyzing checks, can take a while...")

        reports = chain(
            reports,
            check_data_too_old(database, env=environ),
            check_data_not_updated_recently(database, env=environ),
        )

    # --check-empty-databases switch to be on to run "empty_database" (see #146)
    if check_empty_databases is True:
        logger.info("Will analyze databases on this MySQL server, can take a while...")

        reports = chain(
            reports,
            check_empty_database(database),
        )

    return reports


def filter_reports_by_type(reports, checks=None, skip_checks=None):
    """
    :type reports list[indexdigest.utils.LinterEntry]
    :type checks str
    :type skip_checks str
    :rtype: list[indexdigest.utils.LinterEntry]
    """
    if checks:
        return [
            report for report in reports
            if report.linter_type in checks.split(',')
        ]

    if skip_checks:
        return [
            report for report in reports
            if report.linter_type not in skip_checks.split(',')
        ]

    return reports


def filter_reports_by_table(reports, tables=None, skip_tables=None):
    """
    :type reports list[indexdigest.utils.LinterEntry]
    :type tables str
    :type skip_tables str
    :rtype: list[indexdigest.utils.LinterEntry]
    """
    if tables:
        return [
            report for report in reports
            if report.table_name in tables.split(',')
        ]

    if skip_tables:
        return [
            report for report in reports
            if report.table_name not in skip_tables.split(',')
        ]

    return reports


def main():  # pragma: no cover
    """ Main entry point for CLI"""
    logger = logging.getLogger(__name__)

    arguments = docopt(__doc__, version=f'index_digest {get_version()}')
    logger.debug('Options: %s', arguments)

    if 'DSN' not in arguments:
        return

    # connect to the database
    database = Database.connect_dsn(arguments['DSN'])
    logger.debug('Connected to MySQL server v%s', database.get_server_version())

    reports = get_reports(
        database,
        sql_log=arguments.get('--sql-log'),
        analyze_data=arguments.get('--analyze-data'),
        check_empty_databases=arguments.get('--check-empty-databases')
    )

    # handle --checks / --skip-checks
    reports = filter_reports_by_type(
        reports,
        checks=arguments.get('--checks'),
        skip_checks=arguments.get('--skip-checks')
    )

    # handle --tables / --skip-tables
    reports = filter_reports_by_table(
        reports,
        tables=arguments.get('--tables'),
        skip_tables=arguments.get('--skip-tables')
    )

    # handle --format
    formatter = arguments.get('--format') or 'plain'
    logger.info("Using formatter: %s", formatter)

    if formatter == 'plain':
        print(format_plain(database, reports))
    elif formatter == 'syslog':
        ident = getenv('SYSLOG_IDENT', 'index-digest')
        logger.info('Using syslog ident: %s', ident)
        print(format_syslog(database, reports, ident))
    elif formatter == 'yaml':
        print(format_yaml(database, reports))
    else:
        raise IndexDigestError('Unknown formatter provided: {}'.format(formatter))



"""
This linter looks for tables that have really old data
"""
from collections import OrderedDict
from datetime import datetime
from time import time

from indexdigest.utils import LinterEntry, memoize


def get_time_columns(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list
    """
    for table_name in database.get_tables():
        time_columns = [
            column
            for column in database.get_table_columns(table_name)
            if column.is_timestamp_type() or 'time' in column.name
        ]

        # there are no time type columns, skip
        if not time_columns:
            continue

        # for now just check the first time column
        yield (table_name, time_columns[0])


@memoize
def get_boundary_times(database, table_name, column):
    """
    :type database  indexdigest.database.Database
    :type table_name str
    :type column indexdigest.Column.column
    :rtype: dict
    """
    # this may take a while when {column} is not indexed!
    query = 'SELECT /* index-digest */ UNIX_TIMESTAMP(MIN(`{column}`)) as `min`, ' \
            'UNIX_TIMESTAMP(MAX(`{column}`)) as `max` FROM `{table}`'.\
        format(
            column=column.name,
            table=table_name
        )

    timestamps = database.query_dict_row(query)

    # if there's no data in the table, return None
    return timestamps if timestamps.get('min') and timestamps.get('max') else None


def check_data_too_old(database, env=None):
    """
    :type database  indexdigest.database.Database
    :type env dict
    :rtype: list[LinterEntry]
    """
    now = int(time())  # I will probably never understand dates handling in Python

    # set up a diff threshold (in days)
    env = env if env else dict()
    diff_threshold = int(env.get('INDEX_DIGEST_DATA_TOO_OLD_THRESHOLD_DAYS', 3 * 30))

    for (table_name, column) in get_time_columns(database):
        timestamps = get_boundary_times(database, table_name, column)

        if timestamps is None:
            continue

        diff = now - timestamps.get('min')
        # print(table_name, column, timestamps, now, diff)

        if diff > diff_threshold * 86400:
            diff_days = int(diff / 86400)

            metadata = database.get_table_metadata(table_name)

            context = OrderedDict()
            context['diff_days'] = diff_days
            context['data_since'] = str(datetime.fromtimestamp(timestamps.get('min')))
            context['data_until'] = str(datetime.fromtimestamp(timestamps.get('max')))
            context['date_column_name'] = str(column)
            context['schema'] = database.get_table_schema(table_name)
            context['rows'] = database.get_table_rows_estimate(table_name)
            context['table_size_mb'] = \
                1. * (metadata['data_size'] + metadata['index_size']) / 1024 / 1024

            yield LinterEntry(linter_type='data_too_old', table_name=table_name,
                              message='"{}" has rows added {} days ago, '
                                      'consider changing retention policy'.
                              format(table_name, diff_days),
                              context=context)



"""
This linter checks for ...
"""
from collections import OrderedDict

from indexdigest.utils import LinterEntry


def check_use_innodb(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    # in MySQL 8.0 information_schema tables columns are uppercase
    res = database.query_dict_rows("SELECT TABLE_NAME, ENGINE FROM information_schema.tables "
                                   "WHERE ENGINE <> 'InnoDB' and TABLE_SCHEMA = '{}'".
                                   format(database.db_name))

    for row in res:
        context = OrderedDict()
        context['schema'] = database.get_table_schema(row['TABLE_NAME'])
        context['engine'] = row['ENGINE']

        yield LinterEntry(linter_type='use_innodb', table_name=row['TABLE_NAME'],
                          message='"{TABLE_NAME}" uses {ENGINE} storage engine'.
                          format(**row),
                          context=context)



import re
from setuptools import setup, find_packages

# take the version
with open("indexdigest/__init__.py", "r") as fh:
    # e.g. VERSION = '1.5.0'
    last_line = fh.readlines()[-1]
    VERSION = re.search(r'[\d.]+', last_line).group(0)

# @see https://packaging.python.org/tutorials/packaging-projects/#creating-setup-py
with open("README.md", "r") as fh:
    long_description = fh.read()

# @see https://github.com/pypa/sampleproject/blob/master/setup.py
setup(
    name='indexdigest',
    version=VERSION,
    author='Maciej Brencz',
    author_email='maciej.brencz@gmail.com',
    license='MIT',
    description='Analyses your database queries and schema and suggests indices and schema improvements',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/macbre/index-digest',
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Database',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    packages=find_packages(),
    extras_require={
        'dev': [
            'coverage==6.4.4',
            'coveralls==3.3.1',
            'pylint==2.15.2',
            'pytest==7.1.3',
            'pytest-cov==3.0.0',
            'twine==4.0.1',
        ]
    },
    install_requires=[
        'docopt==0.6.2',
        'PyYAML==6.0',
        'mysqlclient==2.1.1',
        'sql_metadata==2.6.0',
        'termcolor==2.0.1',
        'yamlordereddictloader==0.4.0'
    ],
    entry_points={
        'console_scripts': [
            'add_linter=indexdigest.cli.add_linter:main',  # creates a new linter from a template
            'index_digest=indexdigest.cli.script:main',
        ],
    }
)



"""
This linter checks for ...
"""
from indexdigest.utils import LinterEntry

GENERIC_PRIMARY_KEY = 'id'


def check_generic_primary_key(database):
    """
    :type database  indexdigest.database.Database
    :rtype: list[LinterEntry]
    """
    for table_name in database.get_tables():
        indices = [
            index for index in database.get_table_indices(table_name)
            if index.is_primary
        ]

        # no primary index, a different check will take care of it
        if not indices:
            continue

        # there can be only one primary key, take the first one from the list
        primary_key = indices[0]
        # print(table_name, primary_key, primary_key.columns[0])

        if primary_key.columns[0] == GENERIC_PRIMARY_KEY:
            yield LinterEntry(linter_type='generic_primary_key', table_name=table_name,
                              message='"{}" has a primary key called id, '
                                      'use a more meaningful name'.format(table_name),
                              context={"schema": database.get_table_schema(table_name)})
